import matplotlib.pyplot as plt
import numpy as np
import pyfftw as fftw3f
from pyfftw import empty_aligned
from scipy import ndimage
from scipy.fftpack import fftshift, ifftshift, fftn
from skimage import io

NTHREADS = 1

plt.rcParams['figure.figsize'] = [15, 5]
plt.rcParams['image.cmap'] = 'gray'


class PSF:

    def __init__(self, file_path, resolution):
        self.file_path = file_path
        self.resolution = resolution
        self.psf = io.imread(file_path)

    def adjust_resolution(self, img_resolution):
        (pz, px, py) = self.resolution
        (dz, dx, dy) = img_resolution

        self.psf = ndimage.zoom(self.psf, [pz / dz, px / dx, py / dy])

    def resize(self, chunk_size, overlap):
        overlap = tuple(map((2).__mul__, overlap))
        data_size = tuple(map(sum, zip(chunk_size, overlap)))

        if not self.psf.shape == data_size:
            g_ = empty_aligned(shape=data_size, dtype='complex64')
            H_ = empty_aligned(shape=data_size, dtype='complex64')

            sx, sy, sz = np.array(data_size).astype('float') / self.psf.shape
            OT = fftshift(fftn(fftshift(self.psf)))

            if data_size[2] > 1:
                pr = ndimage.zoom(OT.real, [sx, sy, sz], order=1)
                pi = ndimage.zoom(OT.imag, [sx, sy, sz], order=1)
            else:
                pr = ndimage.zoom(OT.real.squeeze(), [sx, sy], order=1).reshape(data_size)
                pi = ndimage.zoom(OT.imag.squeeze(), [sx, sy], order=1).reshape(data_size)

            H_[:] = ifftshift(pr + 1j * pi)
            func = fftw3f.FFTW(H_, g_, direction='FFTW_BACKWARD', axes=range(H_.ndim), threads=NTHREADS,
                               flags=('FFTW_MEASURE',))
            func.execute()
            self.psf = ifftshift(g_.real).clip(min=0)
            self.psf = self.psf / self.psf.sum()
        else:
            self.psf = self.psf / self.psf.sum()

    def calculate_otf(self):
        FTshape = [int(self.psf.shape[0]), int(self.psf.shape[1]), int(self.psf.shape[2] / 2 + 1)]

        self.psf = self.psf.astype('f4')
        psf2 = 1.0 * self.psf[::-1, ::-1, ::-1]

        # allocate memory
        H = empty_aligned(shape=FTshape, dtype='complex64')
        Ht = empty_aligned(shape=FTshape, dtype='complex64')

        # create plans & calculate OTF and conjugate transformed OTF
        p1 = fftw3f.FFTW(self.psf, H, direction='FFTW_FORWARD', axes=range(self.psf.ndim), flags=('FFTW_MEASURE',),
                         threads=NTHREADS)
        p1.execute()
        p2 = fftw3f.FFTW(psf2, Ht, direction='FFTW_FORWARD', axes=range(psf2.ndim), flags=('FFTW_MEASURE',),
                         threads=NTHREADS)
        p2.execute()
        Ht /= self.psf.size
        H /= self.psf.size

        return H, Ht

    def flip(self):
        self.psf = np.flip(self.psf)

    def visualize(self, title="PSF"):

        max_val = np.max(self.psf)
        max_idx = np.unravel_index(np.argmax(self.psf), self.psf.shape)

        fig = plt.figure()
        a = fig.add_subplot(1, 3, 1)
        plt.imshow(self.psf[:, :, max_idx[2]], clim=(0, max_val))
        a.set_title('Y =' + str(max_idx[2]))

        a = fig.add_subplot(1, 3, 2)
        plt.imshow(self.psf[:, max_idx[1], :], clim=(0, max_val))
        a.set_title('X =' + str(max_idx[1]))

        a = fig.add_subplot(1, 3, 3)
        plt.imshow(self.psf[max_idx[0], :, :], clim=(0, max_val))
        a.set_title('Z =' + str(max_idx[0]))
        fig.suptitle(title)
        fig.show()

        self.info()

    def info(self):
        print("Shape: ", self.psf.shape)
        print("Max: ", np.max(self.psf))
        print("Min: ", np.min(self.psf))
        print("Mean: ", np.mean(self.psf))



import numpy as np
from scipy.ndimage import zoom
from skimage.exposure import rescale_intensity as ri


def to_float(img):
    img = img.astype('f4')
    return img


def rescale_intensity(img, in_range=(0, 100)):
    return ri(img, in_range=in_range, out_range=(0, 1))


def resize_block(img, scale):
    return zoom(img, scale)


def invert(img):
    return 1 - img


def intensity_norm(img, imin, imax, gamma=1):
    return ((img - imin) / (imax - imin)) ** gamma


def intensity_clip(img, imin, imax):
    return np.clip(img, imin, imax)


def threshold_mean(img, threshold):
    img[img < threshold] = np.nan
    return np.atleast_3d(np.nanmean(img, axis=(0, 2)))



import dask.array as da

from . import richardsonLucy
from .psf import PSF
from ..adjustments.parallel import to_float


def deconvolve(input_img, H, Ht, num_iters):
    input_shape = input_img.shape
    dec = richardsonLucy.dec_conv()
    dec.psf_calc(H, Ht, input_shape)
    return dec.deconv(input_img, lamb=1, num_iters=num_iters, weights=1).reshape(dec.shape)


def deconvolve_single(input_img, img_resolution, psf_invert, psf_path, psf_res, iterations):
    input_shape = input_img.shape

    psf = PSF(psf_path, psf_res)
    if psf_invert:
        psf.flip()
    psf.visualize("PSF extracted")

    psf.adjust_resolution(img_resolution)
    psf.resize(input_shape, [0, 0, 0])

    H, Ht = psf.calculate_otf()
    print("OTF calculated")

    dec = richardsonLucy.dec_conv()
    dec.psf_calc(H, Ht, input_shape)
    return dec.deconv(input_img, lamb=1, num_iters=iterations, weights=1).reshape(dec.shape)


def deconvolve_parallel(dask_img, img_resolution, chunk_size, overlap, psf_invert, psf_path, psf_res, iterations):
    psf = PSF(psf_path, psf_res)
    if psf_invert:
        psf.flip()

    psf.adjust_resolution(img_resolution)
    psf.resize(chunk_size, overlap)

    H, Ht = psf.calculate_otf()
    print("\t\tOTF calculated")

    dask_img = to_float(dask_img)
    dask_img = da.overlap.overlap(dask_img, depth={0: overlap[0], 1: overlap[1], 2: overlap[2]},
                                  boundary='reflect')
    dask_img = da.map_blocks(deconvolve, dask_img, H, Ht, iterations, dtype='f4')
    return da.overlap.trim_internal(dask_img, {0: overlap[0], 1: overlap[1], 2: overlap[2]})



#!/usr/bin/python

###############
# fftwWisdom.py
#
# Copyright David Baddeley, 2012
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################
import os
import pickle

import pyfftw

WISDOMFILE = os.path.join(os.path.split(__file__)[0], 'fftw_wisdom.pkl')


def load_wisdom():
    if os.path.exists(WISDOMFILE):
        with open(WISDOMFILE, 'rb') as f:
            pyfftw.import_wisdom(pickle.load(f))


def save_wisdom():
    with open(WISDOMFILE, 'wb') as f:
        pickle.dump(pyfftw.export_wisdom(), f)



import glob
import json
import math
import os

import dask.array as da
import pkg_resources
import zarr


def read_config(config_path, session_to_process=None, lasers_to_process=None, pzf_dirs_to_process=None):
    f = open(config_path, "r")
    config = json.loads(f.read())
    f.close()

    root_dir = os.path.dirname(config_path)
    no_of_sessions = len(config['Sessions'])

    stp = range(*(session_format(session_to_process)).indices(no_of_sessions))

    sessions = {}
    idx = 0
    for s in stp:
        sessions[idx] = {}
        print(f'Imaging session: {s}')
        session = config['Sessions'][s]
        s_id = session['Session']
        s_dir = os.path.join(root_dir, s_id)

        lasers = session['Laser Sequence']
        print(f'\tFound laser sequences: {lasers}')

        ll = laser_format(lasers_to_process)
        ltp = ll if ll else lasers

        sessions[idx]['Files'] = {}
        for l in ltp:
            files = sorted(glob.glob(os.path.join(s_dir, f"*{l}*")))
            pzf_dir = list(filter(os.path.isdir, files))
            print(f'\t{l}: {len(pzf_dir)} Blocks with {len(os.listdir(pzf_dir[0]))} planes')
            sessions[s]['Files'][l] = pzf_dir[pzf_dir_format(pzf_dirs_to_process)]

        sessions[idx]['Image Start'] = session['YScan']['Image Start']
        sessions[idx]['Image End'] = session['YScan']['Image End']
        sessions[idx]['Pixel Size'] = session['YScan']['Pixel Size']
        sessions[idx]['Scale Factor'] = session['YScan']['Scale Factor']

        is_reversed = session['ZScan']['Z Increment'] < 0

        img_res_y = 1e6 * session['YScan']['Pixel Size'] * session['YScan']['Scale Factor']
        img_res_z = 1e6 * session['ZScan']['Z Increment']

        sessions[idx]['Image Resolution'] = (abs(img_res_z), 0.23325, img_res_y)
        sessions[idx]['Reversed'] = is_reversed

        overlap = round(2048 - abs(1e6 * session['XScan']['X Increment']) / 0.23325)
        sessions[idx]['Overlap'] = overlap
        idx += 1

    return sessions


def get_files(dirctory, z_range=''):
    block = {'h5_filepath': os.path.join(os.path.dirname(dirctory), 'h5', f'{os.path.basename(dirctory)}.h5')}

    if z_range == '':
        zr = slice(None)
    elif '-' in z_range:
        zr = slice(*map(int, z_range.split('-')))
    else:
        print('Invalid z range')
        return 0

    block['pzf_files'] = sorted(glob.glob(os.path.join(dirctory, "*.pzf")))[zr]
    return block


def input_format(string):
    if '-' in string:
        return slice(*map(int, string.split('-')))
    else:
        return int(string)


def opt_chunksize(depth, xy_chunks=(256, 256), lower=200, upper=400):
    if depth < lower:
        chunk_d = depth
    else:
        divisor = 1
        chunk_d = depth
        while not (lower <= chunk_d < upper):
            divisor += 1
            chunk_d = math.ceil(depth / divisor)

    return (chunk_d,) + xy_chunks


def session_format(string):
    if string == "":
        return slice(None)
    elif '-' in string:
        return slice(*map(int, string.split('-')))
    else:
        num = int(string)
        return slice(num, num + 1, 1)


def laser_format(string):
    if string == "":
        return 0
    elif ',' in string:
        return string.split(',')
    else:
        return [string]


def pzf_dir_format(string):
    if string == "":
        return slice(None)
    elif '-' in string:
        return slice(*map(int, string.split('-')))
    else:
        num = int(string)
        return slice(num, num + 1, 1)


def get_save_path(project_name="CV"):
    path = pkg_resources.resource_filename("musictune", 'data/save_path.json')
    f = open(path, "r")
    paths = json.loads(f.read())
    f.close()

    return paths[project_name]


def to_zarr(dask_img, prefix, chunk_size=None, save_path=None, dtype=None):
    if chunk_size is None:
        chunk_size = dask_img.chunksize
    if dtype is None:
        dtype = dask_img.dtype

    shape = tuple(map(lambda x, c: (math.ceil(x / c) * c), dask_img.shape, chunk_size))

    store_save = zarr.NestedDirectoryStore(save_path)
    zarr_out = zarr.create(shape, chunks=chunk_size, store=store_save, dtype=dtype, fill_value=0,
                           overwrite=True)

    da.to_zarr(dask_img, zarr_out)
    return save_path


def from_zarr(file_path, chunk_size=None):
    zarr_img = zarr.open(zarr.storage.NestedDirectoryStore(file_path), mode='r')

    if chunk_size is None:
        chunk_size = zarr_img.chunks

    return da.from_zarr(zarr_img, chunks=chunk_size)



import numpy as np
import pyfftw as fftw3f
from pyfftw import empty_aligned
from scipy.fftpack import ifftshift

NTHREADS = 4
print('No of threads:', NTHREADS)
from . import fftwWisdom

fftwWisdom.load_wisdom()


class DeconvMappingBase(object):
    """
    Base class for different types of deconvolution. Ultimate deconvolution classes will inherit from both this and a
    a method class (e.g. ICTM Deconvolution)

    Provides implementations of the following:

    AFunc - the forward mapping (computes Af)
    AHFunc - conjugate transpose of forward mapping (computes \bar{A}^T f)
    LFunc - the likelihood function
    LHFunc - conj. transpose of likelihood function

    Also, defines:

    psf_calc - does any pre-computation to get the PSF into a usable form
    prep - allocate memory / setup FFTW plans etc ...

    """

    def Afunc(self, f):
        """ The forward mapping"""
        raise NotImplementedError('Must be over-ridden in derived class')

    def Ahfunc(self, f):
        """ The conjugate transpose of the forward mapping"""
        raise NotImplementedError('Must be over-ridden in derived class')

    def Lfunc(self, f):
        """ The likelihood function (ICTM deconvolution only)"""
        raise NotImplementedError('Must be over-ridden in derived class')

    def Lhfunc(self, f):
        """ The gonjugate transpose of likelihood function (ICTM deconvolution only)"""
        raise NotImplementedError('Must be over-ridden in derived class')

    def psf_calc(self, H, Ht, data_size):
        """
        do any pre-computation on the PSF. e.g. resizing to match the data shape and/or pre-calculating and
        storing the OTF
        """
        raise NotImplementedError('Must be over-ridden in derived class')

    def prep(self):
        """
        Allocate memory and compute FFTW plans etc (this is separate from psf_calc as an aid to distributing Deconvolution
        objects for parallel processing - psf_calc gets called before Deconv objects are passed, prep gets called when
        the deconvolution is run.

        """


class ClassicMappingFFTW(DeconvMappingBase):
    """Classical deconvolution with a stationary PSF using FFTW for convolutions"""

    def prep(self):
        """Precalculate the OTF etc..."""
        # allocate memory
        self._F = empty_aligned(shape=self.FTshape, dtype='complex64')
        self._r = empty_aligned(shape=self.shape, dtype='f4')

        print('\t\tCreating plans for FFTs - this might take a while')
        # calculate plans for other ffts
        self._plan_r_F = fftw3f.FFTW(self._r, self._F, direction='FFTW_FORWARD', axes=range(self._r.ndim),
                                     flags=('FFTW_MEASURE',), threads=NTHREADS)
        self._plan_r_F.execute()
        self._plan_F_r = fftw3f.FFTW(self._F, self._r, direction='FFTW_BACKWARD', axes=range(self._F.ndim),
                                     flags=('FFTW_MEASURE',), threads=NTHREADS)
        self._plan_F_r.execute()
        fftwWisdom.save_wisdom()
        print('Done planning')

    def psf_calc(self, H, Ht, data_size):
        """Precalculate the OTF etc..."""

        self.shape = data_size
        self.FTshape = [int(self.shape[0]), int(self.shape[1]), int(self.shape[2] / 2 + 1)]
        self.Ht = Ht
        self.H = H

    def Lfunc(self, f):
        """convolve with an approximate 2nd derivative likelihood operator in 3D.
        i.e. [[[0,0,0][0,1,0][0,0,0]],[[0,1,0][1,-6,1][0,1,0]],[[0,0,0][0,1,0][0,0,0]]]
        """
        # make our data 3D again
        fs = np.reshape(f, (self.height, self.width, self.depth))
        a = -6 * fs

        a[:, :, 0:-1] += fs[:, :, 1:]
        a[:, :, 1:] += fs[:, :, 0:-1]

        a[:, 0:-1, :] += fs[:, 1:, :]
        a[:, 1:, :] += fs[:, 0:-1, :]

        a[0:-1, :, :] += fs[1:, :, :]
        a[1:, :, :] += fs[0:-1, :, :]

        # flatten data again
        return np.ravel(np.cast['f'](a))

    Lhfunc = Lfunc

    def Afunc(self, f):
        """Forward transform - convolve with the PSF"""
        # fs = reshape(f, (self.height, self.width, self.depth))
        self._r[:] = np.reshape(f, (self._r.shape))
        self._plan_r_F.execute()
        self._F *= self.H
        self._plan_F_r.execute()
        return np.ravel(ifftshift(self._r))

    def Ahfunc(self, f):
        """Conjugate transform - convolve with conj. PSF"""
        self._r[:] = np.reshape(f, (self._r.shape))
        self._plan_r_F.execute()
        self._F *= self.Ht
        self._plan_F_r.execute()
        return np.ravel(ifftshift(self._r))



# -*- coding: utf-8 -*-
"""
Defines a 'wire' format for transmitting or saving image frame data, optionally with Huffman compression and/or
sqrt quantization. The combination of quantization and Huffman coding allows compression ratios of 6-10 fold on typical
microscopy data. Using compression or quantization requires the pyme-compress companion library to be installed which
has optimized c code for performing the compression and quantization. If pyme-compress is compiled and installed on an
AVX capable processor, a throughput in excess of 800MB/s can be achieved.

Most users will just want the :func:`dumps` and :func:`loads` functions
"""

import numpy as np
import six
import zstandard as zstd

if six.PY2:
    _ord = ord
else:
    _ord = lambda a: a

from multiprocessing.pool import ThreadPool

import threading
import logging

logging.basicConfig(level=logging.DEBUG)

try:
    from pymecompress import bcl
except ImportError:
    logging.warning('''Could not import pymecompress library - saving or loading compressed PZF will fail
    (the library is installable from david_baddely conda channel, but requires an AVX capable processor)''')

NUM_COMP_THREADS = 2  # cpu_count()

compPool = ThreadPool(NUM_COMP_THREADS)


def ChunkedHuffmanCompress(data, quantization=None):
    num_chunks = NUM_COMP_THREADS

    chunk_size = int(np.ceil(float(len(data)) / num_chunks))
    raw_chunks = [data[j * chunk_size:(j + 1) * chunk_size].data for j in range(num_chunks)]

    # compPool = ThreadPool(NUM_COMP_THREADS)

    comp_chunk_d = {}

    if quantization is None:
        def _compChunk(c, j):
            comp_chunk_d[j] = bcl.HuffmanCompress(c)

        threads = [threading.Thread(target=_compChunk, args=(rc, j)) for j, rc in enumerate(raw_chunks)]
    else:
        def _compChunk(c, j):
            comp_chunk_d[j] = bcl.HuffmanCompressQuant(c, *quantization)

        threads = [threading.Thread(target=_compChunk, args=(rc, j)) for j, rc in enumerate(raw_chunks)]

    for p in threads:
        # print p
        p.start()

    for p in threads:
        p.join()

    # comp_chunks = compPool.map(bcl.HuffmanCompress, raw_chunks)

    s = np.array([num_chunks], 'u2').tostring()

    for j, r in enumerate(raw_chunks):
        c = comp_chunk_d[j]
        s += np.array([len(c), len(r)], 'u4').tostring()
        s += c.tostring()

    return s


def ChunkedHuffmanCompress_o(data):
    num_chunks = NUM_COMP_THREADS

    chunk_size = int(np.ceil(float(len(data)) / num_chunks))
    raw_chunks = [data[j * chunk_size:(j + 1) * chunk_size].data for j in range(num_chunks)]

    # compPool = ThreadPool(NUM_COMP_THREADS)

    comp_chunks = compPool.map(bcl.HuffmanCompress, raw_chunks)

    s = np.array([num_chunks], 'u2').tostring()

    for c, r in zip(comp_chunks, raw_chunks):
        s += np.array([len(c), len(r)], 'u4').tostring()
        s += c.tostring()

    return s


def _chunkDecompress(args):
    chunk, length = args
    return bcl.HuffmanDecompress(np.fromstring(chunk, 'u1'), length)


def ChunkedHuffmanDecompress(datastring):
    num_chunks = np.fromstring(datastring[:2], 'u2')

    # compPool = ThreadPool(NUM_COMP_THREADS)

    sp = 2

    comp_chunks = []
    for i in range(num_chunks):
        chunk_len, raw_len = np.fromstring(datastring[sp:(sp + 8)], 'u4')
        sp += 8
        comp_chunks.append((datastring[sp:(sp + chunk_len)], raw_len))
        sp += chunk_len

    decomp_chunks = compPool.map(_chunkDecompress, comp_chunks)

    data = np.hstack(decomp_chunks)

    # print data.shape #, comp_chunks
    return data


def ZstdDecompress(data_s, data_format):
    dctx = zstd.ZstdDecompressor()
    data = dctx.decompress(data_s)
    return np.frombuffer(data, dtype=data_format)


FILE_FORMAT_ID = b'BD'
FORMAT_VERSION = 3

DATA_FMT_UINT8 = 0
DATA_FMT_UINT16 = 1
DATA_FMT_FLOAT32 = 2

DATA_FMTS = ['u1', 'u2', 'f4']
DATA_FMTS_SIZES = [1, 2, 4]

DATA_COMP_RAW = 0
DATA_COMP_HUFFCODE = 1
DATA_COMP_HUFFCODE_CHUNKS = 2
DATA_COMP_ZSTD = 3
DATA_COMP_ZSTD_CHUNKS = 4

DATA_QUANT_NONE = 0
DATA_QUANT_SQRT = 1

# header_dtype = [('ID', 'S2'), ('Version', 'u1') , ('DataFormat', 'u1'), ('DataCompression', 'u1'), ('RESERVED0', 'S3'), ('SequenceID', 'i8'),
#                ('FrameNum', 'u4'), ('Width', 'u4'), ('Height', 'u4'), ('Depth', 'u4'),
#                ('FrameTimestamp', 'u8'), ('RESERVED1', 'u8')]

header_dtype = [('ID', 'S2'), ('Version', 'u1'), ('DataFormat', 'u1'), ('DataCompression', 'u1'),
                ('DataQuantization', 'u1'), ('DimOrder', 'S1'), ('RESERVED0', 'S1'), ('SequenceID', 'i8'),
                ('FrameNum', 'u4'), ('Width', 'u4'), ('Height', 'u4'), ('Depth', 'u4'),
                ('FrameTimestamp', 'u8'), ('QuantOffset', 'f4'), ('QuantScale', 'f4')]

# v3 increases header size to support arbitrary offsets to data so that data can be aligned.
header_dtype_v3 = [('ID', 'S2'), ('Version', 'u1'), ('DataFormat', 'u1'), ('DataCompression', 'u1'),
                   ('DataQuantization', 'u1'), ('DimOrder', 'S1'), ('RESERVED0', 'S1'), ('SequenceID', 'i8'),
                   ('FrameNum', 'u4'), ('Width', 'u4'), ('Height', 'u4'), ('Depth', 'u4'),
                   ('FrameTimestamp', 'u8'), ('QuantOffset', 'f4'), ('QuantScale', 'f4'), ('DataOffset', 'u4'),
                   ('RESERVED1', 'S12')]

"""
numpy dtype used to define the file header struct.

Most of the entries should be fairly self explanatory, with the following
deserving a bit more explanation:

:ID: a 2-character string that we can test to see if the file type is consistent
:Version: the version of this format the file uses
:DataFormat: what the data type of individual pixels is
:DataCompression: whether the data is compressed, and which algorithm is used
:SequenceID: A unique identifier for the sequence to which this frame belongs.
    The most important property of this number is that it is unique to
    each sequence. A reasonable method of generation would be to use
    a unix-format integer timestamp for the first dword, and a random
    integer for the second. A hash of the first n image pixels could
    also be used.
:FrameNum: The position of this frame within the sequence
:FrameTimestamp: Space to save camera derived frame timestamps, if available
:Depth: As envisaged, the format is expected to contain individual 2D frames, with
    multiple frames being pulled together in a higher level container to
    construct a sequence or stack. Depth is included just because it doesn't
    take a significant ammount of extra space, but gives us flexibility for
    the future.
"""

HEADER_LENGTH = np.zeros(1, header_dtype).nbytes

HEADER_LENGTH_V3 = np.zeros(1, header_dtype_v3).nbytes


def dumps(data, sequenceID=0, frameNum=0, frameTimestamp=0, compression=DATA_COMP_RAW, quantization=DATA_QUANT_NONE,
          quantizationOffset=0, quantizationScale=1):
    """Dump an image frame (supplied as a numpy array) into a string in PZF format.

    Parameters
    ==========

    data:  ndarray
            The frame as a 2D (or optionally 3D) numpy array

    sequenceID:  int
            A unique identifier for the sequence to which this frame belongs.
            This will let us connect the frame with it's metadata even if
            they end up in different directories etc ...

    frameNum:   int
            The position of this frame within the sequence

    frameTimestamp:  float
            A timestamp for the frame (if provided by the camera)

    compression:  int (enum)
            compression method to use - one of: `PZFFormat.DATA_COMP_RAW`,
            `PZFFormat.DATA_COMP_HUFFCODE`, or `PZFFormat.DATA_COMP_HUFFCODE_CHUNKS`
            Where raw stores the data with no compression, huffcode uses
            Huffman coding, and huffcode chunks breaks the data into chunks
            first, with each chunk meing encodes by a separate thread.

    quantization: int (enum)
            Whether or not the data is quantized before saving.
            One of `DATA_QUANT_NONE` or `DATA_QUANT_SQRT`. If `DATA_QUANT_SQRT`
            is selected, then the data is quantized as follows prior to
            compression:

            .. math:: data_{quant} =  \\frac{\\sqrt{data - quantizationOffset}}{quantizationScale}
    """

    header = np.zeros(1, header_dtype_v3)

    header['ID'] = FILE_FORMAT_ID
    header['Version'] = FORMAT_VERSION

    header['FrameNum'] = frameNum
    header['SequenceID'] = sequenceID
    header['FrameTimestamp'] = frameTimestamp

    header['DataOffset'] = HEADER_LENGTH_V3  # don't support padding on save yet

    if data.dtype == 'uint8':
        header['DataFormat'] = DATA_FMT_UINT8
    elif data.dtype == 'uint16':
        header['DataFormat'] = DATA_FMT_UINT16
    elif data.dtype == 'float32':
        header['DataFormat'] = DATA_FMT_FLOAT32
    else:
        raise RuntimeError('Unsupported data type')

    header['DimOrder'] = 'C'
    header['Width'] = data.shape[0]
    header['Height'] = data.shape[1]
    if data.ndim > 2:
        header['Depth'] = data.shape[2]
    else:
        header['Depth'] = 1

    if quantization == DATA_QUANT_SQRT:
        # sqrt- quantize the data
        header['DataQuantization'] = DATA_QUANT_SQRT
        header['QuantOffset'] = quantizationOffset
        header['QuantScale'] = quantizationScale

    #   qs = 1.0/quantizationScale
    #   data = (np.sqrt(np.maximum(data-quantizationOffset,0))*qs).astype('uint8')

    if data.flags['F_CONTIGUOUS']:
        header['DimOrder'] = 'F'
        d1 = data
        # d1 = np.frombuffer(data, data.dtype) #this will flatten without respecting order - we re-order later
    else:
        # if the array is c-contiguous this will just pass through. If it is non contiguous (i.e. a wierd slice)
        # it will force it to be c-contiguous
        d1 = np.ascontiguousarray(data)

    if compression == DATA_COMP_HUFFCODE:
        header['DataCompression'] = DATA_COMP_HUFFCODE

        if quantization:
            dataString = bcl.HuffmanCompressQuant(d1, quantizationOffset, quantizationScale).tostring()
        else:
            d2 = bcl.HuffmanCompress(d1)
            dataString = d2.tostring()
    elif compression == DATA_COMP_HUFFCODE_CHUNKS:
        header['DataCompression'] = DATA_COMP_HUFFCODE_CHUNKS

        dataString = ChunkedHuffmanCompress(d1)
    else:
        # print('saving raw')
        # print(header['DimOrder'][0])
        dataString = d1.tostring(order=header['DimOrder'][0])

    return header.tostring() + dataString


def load_header(datastring):
    if (_ord(datastring[2]) >= 3):
        return np.fromstring(datastring[:HEADER_LENGTH_V3], header_dtype_v3)
    else:
        return np.fromstring(datastring[:HEADER_LENGTH], header_dtype)


def display_header(datastring):
    header = load_header(datastring)

    if header['Version'] >= 2:
        dim_order = header['DimOrder'][0]
    else:
        dim_order = 'C'

    w, h, d = header['Width'][0], header['Height'][0], header['Depth'][0]

    if header['Version'] < 3:
        data_offset = HEADER_LENGTH
    else:
        data_offset = int(header['DataOffset'])

    if header['DataCompression'] == DATA_COMP_RAW:
        compression = 'None'
    elif header['DataCompression'] == DATA_COMP_HUFFCODE:
        compression = 'Huffmancode'
    elif header['DataCompression'] == DATA_COMP_HUFFCODE_CHUNKS:
        compression = 'Chunked Huffmancode'
    elif header['DataCompression'] == DATA_COMP_ZSTD:
        compression = 'Zstd'
    elif header['DataCompression'] == DATA_COMP_ZSTD_CHUNKS:
        compression = 'Chunked Zstd'
    else:
        raise RuntimeError('Compression type not understood')

    header_json = {
        'ID': str(header['ID'][0]),
        'Version': header['Version'][0],
        'Format': DATA_FMTS[header['DataFormat'][0]],
        'Compression': compression,
        'Quantisation': 'SQRT' if header['DataQuantization'] else 'None',
        'Order': str(dim_order),
        'SequenceId': header['SequenceID'][0],
        'Frame no': header['FrameNum'][0],
        'Width': w,
        'Height': h,
        'Depth': d,
        'Data offset': data_offset,
    }

    return header_json


def loads(datastring):
    """
    Loads image data from a string in PZF format.

    Parameters
    ----------
    datastring : string / bytes
        The encoded data

    Returns
    -------

    data : ndarray
        The image data as a numpy array

    header : recarray
        The image header, as a numpy record array with the :const:`header_dtype` dtype.

    """
    header = load_header(datastring)

    if not header['ID'] == FILE_FORMAT_ID:
        raise RuntimeError("Invalid format: This doesn't appear to be a PZF file")

    if header['Version'] >= 2:
        dimOrder = header['DimOrder'][0]
    else:
        dimOrder = 'C'

    # print(dimOrder)

    w, h, d = header['Width'][0], header['Height'][0], header['Depth'][0]

    if header['DataQuantization'] == DATA_QUANT_SQRT:
        # quantized data is always 8 bit
        outsize = w * h * d
    else:
        outsize = w * h * d * DATA_FMTS_SIZES[int(header['DataFormat'])]

    if header['Version'] < 3:
        data_offset = HEADER_LENGTH
    else:
        data_offset = int(header['DataOffset'])

    data_s = datastring[data_offset:]
    data_format = DATA_FMTS[header['DataFormat'][0]]

    # logging.debug('About to decompress')
    # logging.debug({k:header[0][k] for k in header.dtype.names})

    # logging.debug('Compressed size: %s' % len(data_s))

    if header['DataCompression'] == DATA_COMP_RAW:
        # no need to decompress
        data = np.fromstring(data_s, 'u1')
    elif header['DataCompression'] == DATA_COMP_HUFFCODE:
        # logging.debug('Decompressing ...')
        data = bcl.HuffmanDecompress(np.fromstring(data_s, 'u1'), outsize)
    elif header['DataCompression'] == DATA_COMP_HUFFCODE_CHUNKS:
        data = ChunkedHuffmanDecompress(data_s)
    elif header['DataCompression'] == DATA_COMP_ZSTD:
        data = ZstdDecompress(data_s, data_format)
    else:
        raise RuntimeError('Compression type not understood')

    # logging.debug('Uncompressed shape: %s, %s, (%d, %d, %d)' % (data.shape, w * h * d, w, h, d))

    if header['DataQuantization'] == DATA_QUANT_SQRT:
        # un-quantize data
        # logging.debug('Dequantizing')
        #         print(data.max())
        #         data = data.astype('f') * header['QuantScale']
        data = data.astype('f') * 1.0
        #         print('data dtype: %s' % data.dtype)
        #         print('data shape: %s' % data.shape)
        data = data * data
    #         data = (data * data + header['QuantOffset'])

    #     print(dimOrder, [w, h, d])
    data = data.reshape([d, h, w], order=dimOrder).T

    return data, header



import dask
import numpy as np

from ..io import PZFFormat


@dask.delayed
def decompress(pzf_filepath):
    with open(pzf_filepath, 'rb') as f:
        datastring = f.read()

    data, header = PZFFormat.loads(datastring)
    return data


def merge_lines(block, weighted_pixels, option="merge", start_index=None, crop_length=None, line_no=4):
    if option == "merge":
        img_cropped = block[:, :, start_index:start_index + crop_length, :]
        return np.sum(np.expand_dims(weighted_pixels.T, (0, 3)) * img_cropped, 2, keepdims=True)
    elif option == "single":
        return block[:, :, line_no:line_no + 1, :]


def reposition_lines(data, sample_positions, grid):
    sample_positions = np.squeeze(sample_positions)
    data = np.squeeze(data)
    grid_img = np.zeros_like(data)
    x_length = sample_positions.shape[0]
    sample_idx = 0

    if sample_positions[-1] < sample_positions[0]:
        sample_positions = np.flip(sample_positions)
        data = np.flip(data, 1)

    for i in range(x_length):

        if grid[i] < sample_positions[0]:
            grid_img[:, i] = data[:, 0]
        elif grid[i] > sample_positions[-1]:
            grid_img[:, i] = data[:, -1]
        else:
            while not (sample_positions[sample_idx] <= grid[i] <= sample_positions[sample_idx + 1]):
                sample_idx += 1

            right = sample_positions[sample_idx + 1] - grid[i]
            left = grid[i] - sample_positions[sample_idx]
            grid_img[:, i] = ((right * data[:, sample_idx]) + (left * data[:, sample_idx + 1])) / (right + left)

    return np.expand_dims(grid_img, 0)



import json
import math
import os
import shutil
import sys
import time
from ast import literal_eval

import dask.array as da
import h5py as h5
import numpy as np
import pkg_resources
import zarr
from distributed import Client
from musictune.deconvolution.deconvolve import deconvolve_parallel
from musictune.io.modules import decompress, merge_lines, reposition_lines
from musictune.io.utilities import read_config, get_save_path, get_files, opt_chunksize, from_zarr, to_zarr

f = open(sys.argv[1], "r")
param = json.loads(f.read())
f.close()

config_path = param['config_path']
f = open(config_path, "r")
config = json.loads(f.read())
f.close()

# directories and range options
s = param['range']['sessions']
l = param['range']['lasers']
d = param['range']['pzf_dirs']

sessions = read_config(config_path, s, l, d)

z_range = str(param['range']['z_range'])

# Line weights and paramters
line_weights_file = param['lines']['weights_file']
option = param['lines']['weights_option']
line_no = param['lines']['single_line_no']

# Deconvolution parameters
deconv_status = literal_eval(param['deconvolve']['status'])
overlap = literal_eval(param['deconvolve']['overlap'])
psf_path = param['deconvolve']['psf_path']
psf_res = param['deconvolve']['psf_res']
iterations = int(param['deconvolve']['iterations'])
td = param['save']['tmp_dir']

tmp_chunks = literal_eval(param['save']['tmp_chunks'])
xy_chunks = literal_eval(param['save']['xy_chunks'])
zchunk_range = [*map(int, param['save']['zchunk_range'].split('-'))]

n_workers = param['cluster']['n_workers']
threads_per_worker = param['cluster']['threads_per_worker']
ld = param['cluster']['local_directory']

local_directory = os.environ.get(ld)
if not local_directory:
    local_directory = ld

tmp_dir = os.environ.get(td)
if not tmp_dir:
    tmp_dir = td

if line_weights_file == "":
    f = open(pkg_resources.resource_filename("musictune", 'data/Weights.json'), "r")
    weights = json.loads(f.read())
    f.close()
else:
    f = open(line_weights_file, "r")
    weights = json.loads(f.read())
    f.close()

if psf_path == "measured":
    psf_path = pkg_resources.resource_filename("musictune", 'data/PSF/CV_us_640_fine.tif')
    psf_res = (0.23325, 0.23325, 0.23325)
elif psf_path == "theoretical":
    psf_path = pkg_resources.resource_filename("musictune", 'data/PSF/PSF_16bit_full.tif')
    psf_res = (0.933, 0.933, 0.933)

# if __name__ == "__main__":
#     client = Client(processes=False, n_workers=n_workers, threads_per_worker=threads_per_worker,
#                     local_directory=local_directory)

client = Client('tcp://127.0.0.1:'+sys.argv[2])
print(client)

start_time = time.time()

start_index = weights['ReadCrop']['index']
crop_length = weights['ReadCrop']['length']
weight = np.array(weights['Weight'])
weighted_pixels = np.expand_dims(weight, 1) * np.array(weights['Weighted Pixels'])

save_root = os.path.join(get_save_path(config['Basename']["Project"]), config['Basename']["Sample"])
save_dir = os.path.join(save_root, config['Basename']["Sequence"])

for sess in sessions:
    print(f"Processing session: {sess}")
    img_res = sessions[sess]['Image Resolution']
    print(f"Image resolution: {img_res}")
    psf_invert = not sessions[sess]['Reversed']

    lasers = sessions[sess]['Files']
    grid = 1e8 * np.arange(sessions[sess]['Image Start'], sessions[sess]['Image End'],
                           sessions[sess]['Pixel Size'] * sessions[sess]['Scale Factor'])

    for las in lasers:
        print(f"\tProcessing Lasers: {las}")
        directories = lasers[las]

        for pzf_dir in directories:
            print(f"\t\tProcessing PZF directory: {pzf_dir}")

            block = get_files(pzf_dir, z_range=z_range)

            sample = decompress(block['pzf_files'][0]).compute()

            delayed_planes = [decompress(fn) for fn in block['pzf_files']]
            delayed_planes = [da.from_delayed(x, shape=sample.shape, dtype='f4') for x in delayed_planes]

            da_line_planes = da.stack(delayed_planes)

            coordinate_file = h5.File(block['h5_filepath'], 'r')
            coordinates = da.from_array(
                np.expand_dims(np.array([coordinate_file.get(fn.split(os.sep)[-1])[:] for fn in block['pzf_files']]),
                               1), chunks=(1, 1, da_line_planes.chunksize[-1]))

            ## Merge lines
            merge_chunk_size = (1, da_line_planes.chunksize[1], 1, da_line_planes.chunksize[3])

            da_planes = da.map_blocks(merge_lines, da_line_planes, weighted_pixels, option=option,
                                      start_index=start_index,
                                      crop_length=crop_length, line_no=line_no, chunks=merge_chunk_size,
                                      dtype='f4').squeeze()

            da_planes_repositioned = da.map_blocks(reposition_lines, da_planes, coordinates, grid, dtype='f4')

            save_path = os.path.join(save_dir, *pzf_dir.split(os.sep)[-2:]) + '.zarr'
            tmp_save_path = tmp_dir + save_path

            shape = tuple(map(lambda x, c: (math.ceil(x / c) * c), da_planes.shape, tmp_chunks))

            store = zarr.NestedDirectoryStore(tmp_save_path)
            z_out = zarr.create(shape=shape, chunks=tmp_chunks, dtype=da_planes.dtype, store=store, overwrite=True,
                                fill_value=0)

            print(f"\t\tSaving tmp image: {tmp_save_path}")
            tp = da.to_zarr(da_planes, z_out)

            chunk_size = opt_chunksize(da_planes.shape[0], xy_chunks, zchunk_range[0], zchunk_range[1])

            dask_img = from_zarr(tmp_save_path, chunk_size=chunk_size)
            print(f"\t\tRechunking image: {dask_img}")

            if deconv_status:
                print("\t\tDeconvolving")

                dask_img = deconvolve_parallel(dask_img,
                                               img_resolution=img_res,
                                               chunk_size=chunk_size,
                                               overlap=overlap,
                                               psf_invert=psf_invert,
                                               psf_path=psf_path,
                                               psf_res=psf_res,
                                               iterations=iterations)

            path = to_zarr(dask_img, prefix='', chunk_size=chunk_size, save_path=save_path)

            print(f"\t\tSaved:{path}")

            shutil.rmtree(tmp_save_path)
            print("\t\tTmp file removed\n")

elapsed_time = time.time() - start_time
print("Elapsed time:", elapsed_time)

config["ZARR Conversion"] = param
config["ZARR Conversion"]["elapsed_time"] = elapsed_time

with open(os.path.join(save_root, config['Basename']['Sequence'] + '.json'), 'w') as outfile:
    json.dump(config, outfile, indent=4)

client.close()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="musictune",
    version="0.0.3",
    author="vibujithan",
    author_email="author@example.com",
    description="Parallel image processing package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vibujithan/musictune",
    project_urls={
        "Bug Tracker": "https://github.com/vibujithan/musictune/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    package_data={
        "musictune": ["data/**"],
        "": ["examples/**"]
    },
    scripts=['src/musictune/scripts/pzf2zarr.py']
)



#!/usr/bin/python
##################
# richardsonLucy.py
#
# Copyright David Baddeley, 2011
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##################

from . import fftwWisdom, deconv_mappings as dec

fftwWisdom.load_wisdom()
import numpy as np


class RichardsonLucyDeconvolution(object):
    """Deconvolution class, implementing a variant of the Richardson-Lucy algorithm.

    Derived classed should additionally define the following methods:
    AFunc - the forward mapping (computes Af)
    AHFunc - conjugate transpose of forward mapping (computes \bar{A}^T f)
    LFunc - the likelihood function
    LHFunc - conj. transpose of likelihood function

    see dec_conv for an implementation of conventional image deconvolution with a
    measured, spatially invariant PSF
    """

    def __init__(self) -> object:
        pass

    def startGuess(self, data):
        """starting guess for deconvolution - can be overridden in derived classes
        but the data itself is usually a pretty good guess.
        """
        return 0 * data + data.mean()

    def deconvp(self, args):
        """ convenience function for deconvolving in parallel using processing.Pool.map"""
        return self.deconv(*args)
        # return 0

    def deconv(self, data, lamb, num_iters=10, weights=1, bg=0):
        """This is what you actually call to do the deconvolution.
        parameters are:

        data - the raw data
        lamb - the regularisation parameter (ignored - kept for compatibility with ICTM)
        num_iters - number of iterations (note that the convergence is fast when
                    compared to many algorithms - e.g Richardson-Lucy - and the
                    default of 10 will usually already give a reasonable result)


        """
        # remember what shape we are
        self.dataShape = np.shape(data)

        if 'prep' in dir(self) and not '_F' in dir(self):
            self.prep()

        if not np.isscalar(weights):
            self.mask = 1 - weights  # > 0
        else:
            self.mask = 1 - np.isfinite(data.ravel())

        # guess a starting estimate for the object
        self.f = self.startGuess(data).ravel() - bg
        self.fs = np.reshape(self.f, (self.shape))

        # make things 1 dimensional
        # self.f = self.f.ravel()
        data = np.ravel(data)

        self.loopcount = 0

        while self.loopcount < num_iters:
            self.loopcount += 1
            # the residuals
            self.res = weights * (data / (self.Afunc(self.f) + 1e-12 + bg)) + self.mask
            # adjustment
            adjFact = self.Ahfunc(self.res)
            fnew = self.f * adjFact
            # set the current estimate to out new estimate
            self.f[:] = fnew
            # print(100*self.loopcount/num_iters)
        return np.real(self.fs)


class dec_conv(RichardsonLucyDeconvolution, dec.ClassicMappingFFTW):
    def __init__(self, *args, **kwargs):
        RichardsonLucyDeconvolution.__init__(self, *args, **kwargs)
